@@ -3,7 +3,6 @@ use crate::trampoline::MemoryCreatorProxy;
 use anyhow::{bail, Result};
 use serde::{Deserialize, Serialize};
 use std::cmp;
-use std::convert::TryFrom;
 use std::fmt;
 #[cfg(feature = "cache")]
 use std::path::Path;
@@ -11,10 +10,8 @@ use std::sync::Arc;
 use wasmparser::WasmFeatures;
 #[cfg(feature = "cache")]
 use wasmtime_cache::CacheConfig;
-use wasmtime_environ::settings::{self, Configurable, SetError};
-use wasmtime_environ::{isa, isa::TargetIsa, Tunables};
-use wasmtime_jit::{native, CompilationStrategy, Compiler};
-use wasmtime_profiling::{JitDumpAgent, NullProfilerAgent, ProfilingAgent, VTuneAgent};
+use wasmtime_environ::{CompilerBuilder, Tunables};
+use wasmtime_jit::{JitDumpAgent, NullProfilerAgent, ProfilingAgent, VTuneAgent};
 use wasmtime_runtime::{
     InstanceAllocator, OnDemandInstanceAllocator, PoolingInstanceAllocator, RuntimeMemoryCreator,
 };
@@ -136,7 +133,7 @@ pub struct ModuleLimits {
     /// The reservation size of each linear memory is controlled by the
     /// [`static_memory_maximum_size`](Config::static_memory_maximum_size) setting and this value cannot
     /// exceed the configured static memory maximum size.
-    pub memory_pages: u32,
+    pub memory_pages: u64,
 }
 
 impl Default for ModuleLimits {
@@ -333,17 +330,32 @@ impl Default for InstanceAllocationStrategy {
     }
 }
 
+#[derive(Clone)]
+/// Configure the strategy used for versioning in serializing and deserializing [`crate::Module`].
+pub enum ModuleVersionStrategy {
+    /// Use the wasmtime crate's Cargo package version.
+    WasmtimeVersion,
+    /// Use a custom version string. Must be at most 255 bytes.
+    Custom(String),
+    /// Emit no version string in serialization, and accept all version strings in deserialization.
+    None,
+}
+
+impl Default for ModuleVersionStrategy {
+    fn default() -> Self {
+        ModuleVersionStrategy::WasmtimeVersion
+    }
+}
+
 /// Global configuration options used to create an [`Engine`](crate::Engine)
 /// and customize its behavior.
 ///
 /// This structure exposed a builder-like interface and is primarily consumed by
 /// [`Engine::new()`](crate::Engine::new)
-#[derive(Clone)]
 pub struct Config {
-    pub(crate) flags: settings::Builder,
-    pub(crate) isa_flags: isa::Builder,
+    #[cfg(compiler)]
+    pub(crate) compiler: Box<dyn CompilerBuilder>,
     pub(crate) tunables: Tunables,
-    pub(crate) strategy: CompilationStrategy,
     #[cfg(feature = "cache")]
     pub(crate) cache_config: CacheConfig,
     pub(crate) profiler: Arc<dyn ProfilingAgent>,
@@ -355,31 +367,19 @@ pub struct Config {
     #[cfg(feature = "async")]
     pub(crate) async_stack_size: usize,
     pub(crate) async_support: bool,
-    pub(crate) deserialize_check_wasmtime_version: bool,
+    pub(crate) module_version: ModuleVersionStrategy,
+    pub(crate) parallel_compilation: bool,
+    pub(crate) paged_memory_initialization: bool,
 }
 
 impl Config {
     /// Creates a new configuration object with the default configuration
     /// specified.
     pub fn new() -> Self {
-        let mut flags = settings::builder();
-
-        // There are two possible traps for division, and this way
-        // we get the proper one if code traps.
-        flags
-            .enable("avoid_div_traps")
-            .expect("should be valid flag");
-
-        // We don't use probestack as a stack limit mechanism
-        flags
-            .set("enable_probestack", "false")
-            .expect("should be valid flag");
-
         let mut ret = Self {
             tunables: Tunables::default(),
-            flags,
-            isa_flags: native::builder(),
-            strategy: CompilationStrategy::Auto,
+            #[cfg(compiler)]
+            compiler: compiler_builder(Strategy::Auto).unwrap(),
             #[cfg(feature = "cache")]
             cache_config: CacheConfig::new_cache_disabled(),
             profiler: Arc::new(NullProfilerAgent),
@@ -391,10 +391,16 @@ impl Config {
             #[cfg(feature = "async")]
             async_stack_size: 2 << 20,
             async_support: false,
-            deserialize_check_wasmtime_version: true,
+            module_version: ModuleVersionStrategy::default(),
+            parallel_compilation: true,
+            // Default to paged memory initialization when using uffd on linux
+            paged_memory_initialization: cfg!(all(target_os = "linux", feature = "uffd")),
         };
-        ret.cranelift_debug_verifier(false);
-        ret.cranelift_opt_level(OptLevel::Speed);
+        #[cfg(compiler)]
+        {
+            ret.cranelift_debug_verifier(false);
+            ret.cranelift_opt_level(OptLevel::Speed);
+        }
         ret.wasm_reference_types(true);
         ret.wasm_multi_value(true);
         ret.wasm_bulk_memory(true);
@@ -414,11 +420,12 @@ impl Config {
     /// # Errors
     ///
     /// This method will error if the given target triple is not supported.
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn target(&mut self, target: &str) -> Result<&mut Self> {
         use std::str::FromStr;
-        self.isa_flags = native::lookup(
-            target_lexicon::Triple::from_str(target).map_err(|e| anyhow::anyhow!(e))?,
-        )?;
+        self.compiler
+            .target(target_lexicon::Triple::from_str(target).map_err(|e| anyhow::anyhow!(e))?)?;
 
         Ok(self)
     }
@@ -674,9 +681,12 @@ impl Config {
     pub fn wasm_reference_types(&mut self, enable: bool) -> &mut Self {
         self.features.reference_types = enable;
 
-        self.flags
-            .set("enable_safepoints", if enable { "true" } else { "false" })
-            .unwrap();
+        #[cfg(compiler)]
+        {
+            self.compiler
+                .set("enable_safepoints", if enable { "true" } else { "false" })
+                .unwrap();
+        }
 
         // The reference types proposal depends on the bulk memory proposal.
         if enable {
@@ -708,10 +718,13 @@ impl Config {
     /// [proposal]: https://github.com/webassembly/simd
     pub fn wasm_simd(&mut self, enable: bool) -> &mut Self {
         self.features.simd = enable;
-        let val = if enable { "true" } else { "false" };
-        self.flags
-            .set("enable_simd", val)
-            .expect("should be valid flag");
+        #[cfg(compiler)]
+        {
+            let val = if enable { "true" } else { "false" };
+            self.compiler
+                .set("enable_simd", val)
+                .expect("should be valid flag");
+        }
         self
     }
 
@@ -771,6 +784,21 @@ impl Config {
         self
     }
 
+    /// Configures whether the WebAssembly memory64 [proposal] will
+    /// be enabled for compilation.
+    ///
+    /// Note that this the upstream specification is not finalized and Wasmtime
+    /// may also have bugs for this feature since it hasn't been exercised
+    /// much.
+    ///
+    /// This is `false` by default.
+    ///
+    /// [proposal]: https://github.com/webassembly/memory64
+    pub fn wasm_memory64(&mut self, enable: bool) -> &mut Self {
+        self.features.memory64 = enable;
+        self
+    }
+
     /// Configures which compilation strategy will be used for wasm modules.
     ///
     /// This method can be used to configure which compiler is used for wasm
@@ -784,17 +812,10 @@ impl Config {
     /// Some compilation strategies require compile-time options of `wasmtime`
     /// itself to be set, but if they're not set and the strategy is specified
     /// here then an error will be returned.
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn strategy(&mut self, strategy: Strategy) -> Result<&mut Self> {
-        self.strategy = match strategy {
-            Strategy::Auto => CompilationStrategy::Auto,
-            Strategy::Cranelift => CompilationStrategy::Cranelift,
-            #[cfg(feature = "lightbeam")]
-            Strategy::Lightbeam => CompilationStrategy::Lightbeam,
-            #[cfg(not(feature = "lightbeam"))]
-            Strategy::Lightbeam => {
-                anyhow::bail!("lightbeam compilation strategy wasn't enabled at compile time");
-            }
-        };
+        self.compiler = compiler_builder(strategy)?;
         Ok(self)
     }
 
@@ -819,9 +840,11 @@ impl Config {
     /// developers of wasmtime itself.
     ///
     /// The default value for this is `false`
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn cranelift_debug_verifier(&mut self, enable: bool) -> &mut Self {
         let val = if enable { "true" } else { "false" };
-        self.flags
+        self.compiler
             .set("enable_verifier", val)
             .expect("should be valid flag");
         self
@@ -834,13 +857,15 @@ impl Config {
     /// more information see the documentation of [`OptLevel`].
     ///
     /// The default value for this is `OptLevel::None`.
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn cranelift_opt_level(&mut self, level: OptLevel) -> &mut Self {
         let val = match level {
             OptLevel::None => "none",
             OptLevel::Speed => "speed",
             OptLevel::SpeedAndSize => "speed_and_size",
         };
-        self.flags
+        self.compiler
             .set("opt_level", val)
             .expect("should be valid flag");
         self
@@ -854,9 +879,11 @@ impl Config {
     /// This is not required by the WebAssembly spec, so it is not enabled by default.
     ///
     /// The default value for this is `false`
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn cranelift_nan_canonicalization(&mut self, enable: bool) -> &mut Self {
         let val = if enable { "true" } else { "false" };
-        self.flags
+        self.compiler
             .set("enable_nan_canonicalization", val)
             .expect("should be valid flag");
         self
@@ -876,16 +903,10 @@ impl Config {
     /// # Errors
     ///
     /// This method can fail if the flag's name does not exist.
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub unsafe fn cranelift_flag_enable(&mut self, flag: &str) -> Result<&mut Self> {
-        if let Err(err) = self.flags.enable(flag) {
-            match err {
-                SetError::BadName(_) => {
-                    // Try the target-specific flags.
-                    self.isa_flags.enable(flag)?;
-                }
-                _ => bail!(err),
-            }
-        }
+        self.compiler.enable(flag)?;
         Ok(self)
     }
 
@@ -902,16 +923,10 @@ impl Config {
     ///
     /// This method can fail if the flag's name does not exist, or the value is not appropriate for
     /// the flag type.
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub unsafe fn cranelift_flag_set(&mut self, name: &str, value: &str) -> Result<&mut Self> {
-        if let Err(err) = self.flags.set(name, value) {
-            match err {
-                SetError::BadName(_) => {
-                    // Try the target-specific flags.
-                    self.isa_flags.set(name, value)?;
-                }
-                _ => bail!(err),
-            }
-        }
+        self.compiler.set(name, value)?;
         Ok(self)
     }
 
@@ -989,6 +1004,27 @@ impl Config {
         self
     }
 
+    /// Sets whether or not an attempt is made to initialize linear memories by page.
+    ///
+    /// This setting is `false` by default and Wasmtime initializes linear memories
+    /// by copying individual data segments from the compiled module.
+    ///
+    /// Setting this to `true` will cause compilation to attempt to organize the
+    /// data segments into WebAssembly pages and linear memories are initialized by
+    /// copying each page rather than individual data segments.
+    ///
+    /// Modules that import a memory or have data segments that use a global base
+    /// will continue to be initialized by copying each data segment individually.
+    ///
+    /// When combined with the `uffd` feature on Linux, this will allow Wasmtime
+    /// to delay initialization of a linear memory page until it is accessed
+    /// for the first time during WebAssembly execution; this may improve
+    /// instantiation performance as a result.
+    pub fn paged_memory_initialization(&mut self, value: bool) -> &mut Self {
+        self.paged_memory_initialization = value;
+        self
+    }
+
     /// Configures the maximum size, in bytes, where a linear memory is
     /// considered static, above which it'll be considered dynamic.
     ///
@@ -1079,7 +1115,7 @@ impl Config {
     /// pooling allocator.
     pub fn static_memory_maximum_size(&mut self, max_size: u64) -> &mut Self {
         let max_pages = max_size / u64::from(wasmtime_environ::WASM_PAGE_SIZE);
-        self.tunables.static_memory_bound = u32::try_from(max_pages).unwrap_or(u32::max_value());
+        self.tunables.static_memory_bound = max_pages;
         self
     }
 
@@ -1171,6 +1207,45 @@ impl Config {
         self
     }
 
+    /// Configures the size, in bytes, of the extra virtual memory space
+    /// reserved after a "dynamic" memory for growing into.
+    ///
+    /// For the difference between static and dynamic memories, see the
+    /// [`Config::static_memory_maximum_size`]
+    ///
+    /// Dynamic memories can be relocated in the process's virtual address space
+    /// on growth and do not always reserve their entire space up-front. This
+    /// means that a growth of the memory may require movement in the address
+    /// space, which in the worst case can copy a large number of bytes from one
+    /// region to another.
+    ///
+    /// This setting configures how many bytes are reserved after the initial
+    /// reservation for a dynamic memory for growing into. A value of 0 here
+    /// means that no extra bytes are reserved and all calls to `memory.grow`
+    /// will need to relocate the wasm linear memory (copying all the bytes). A
+    /// value of 1 megabyte, however, means that `memory.grow` can allocate up
+    /// to a megabyte of extra memory before the memory needs to be moved in
+    /// linear memory.
+    ///
+    /// Note that this is a currently simple heuristic for optimizing the growth
+    /// of dynamic memories, primarily implemented for the memory64 propsal
+    /// where all memories are currently "dynamic". This is unlikely to be a
+    /// one-size-fits-all style approach and if you're an embedder running into
+    /// issues with dynamic memories and growth and are interested in having
+    /// other growth strategies available here please feel free to [open an
+    /// issue on the Wasmtime repository][issue]!
+    ///
+    /// [issue]: https://github.com/bytecodealliance/wasmtime/issues/ne
+    ///
+    /// ## Default
+    ///
+    /// For 64-bit platforms this defaults to 2GB, and for 32-bit platforms this
+    /// defaults to 1MB.
+    pub fn dynamic_memory_reserved_for_growth(&mut self, reserved: u64) -> &mut Self {
+        self.tunables.dynamic_memory_growth_reserve = round_up_to_pages(reserved);
+        self
+    }
+
     /// Indicates whether a guard region is present before allocations of
     /// linear memory.
     ///
@@ -1196,39 +1271,37 @@ impl Config {
         self
     }
 
-    /// Configure whether deserialized modules should validate version
-    /// information. This only effects [`crate::Module::deserialize()`], which is
-    /// used to load compiled code from trusted sources.  When true,
-    /// [`crate::Module::deserialize()`] verifies that the wasmtime crate's
-    /// `CARGO_PKG_VERSION` matches with the version in the binary, which was
-    /// produced by [`crate::Module::serialize`] or
-    /// [`crate::Engine::precompile_module`].
+    /// Configure the version information used in serialized and deserialzied [`crate::Module`]s.
+    /// This effects the behavior of [`crate::Module::serialize()`], as well as
+    /// [`crate::Module::deserialize()`] and related functions.
     ///
-    /// This value defaults to true.
-    pub fn deserialize_check_wasmtime_version(&mut self, check: bool) -> &mut Self {
-        self.deserialize_check_wasmtime_version = check;
+    /// The default strategy is to use the wasmtime crate's Cargo package version.
+    pub fn module_version(&mut self, strategy: ModuleVersionStrategy) -> Result<&mut Self> {
+        match strategy {
+            // This case requires special precondition for assertion in SerializedModule::to_bytes
+            ModuleVersionStrategy::Custom(ref v) => {
+                if v.as_bytes().len() > 255 {
+                    bail!("custom module version cannot be more than 255 bytes: {}", v);
+                }
+            }
+            _ => {}
+        }
+        self.module_version = strategy;
+        Ok(self)
+    }
+
+    /// Configure wether wasmtime should compile a module using multiple threads.
+    ///
+    /// Disabling this will result in a single thread being used to compile the wasm bytecode.
+    ///
+    /// By default parallel compilation is enabled.
+    #[cfg(feature = "parallel-compilation")]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "parallel-compilation")))]
+    pub fn parallel_compilation(&mut self, parallel: bool) -> &mut Self {
+        self.parallel_compilation = parallel;
         self
     }
 
-    pub(crate) fn target_isa(&self) -> Box<dyn TargetIsa> {
-        self.isa_flags
-            .clone()
-            .finish(settings::Flags::new(self.flags.clone()))
-    }
-
-    pub(crate) fn target_isa_with_reference_types(&self) -> Box<dyn TargetIsa> {
-        let mut flags = self.flags.clone();
-        flags.set("enable_safepoints", "true").unwrap();
-        self.isa_flags.clone().finish(settings::Flags::new(flags))
-    }
-
-    pub(crate) fn build_compiler(&self, allocator: &dyn InstanceAllocator) -> Compiler {
-        let isa = self.target_isa();
-        let mut tunables = self.tunables.clone();
-        allocator.adjust_tunables(&mut tunables);
-        Compiler::new(isa, self.strategy, tunables, self.features)
-    }
-
     pub(crate) fn build_allocator(&self) -> Result<Box<dyn InstanceAllocator>> {
         #[cfg(feature = "async")]
         let stack_size = self.async_stack_size;
@@ -1256,6 +1329,19 @@ impl Config {
     }
 }
 
+#[cfg(compiler)]
+fn compiler_builder(strategy: Strategy) -> Result<Box<dyn CompilerBuilder>> {
+    match strategy {
+        Strategy::Auto | Strategy::Cranelift => Ok(wasmtime_cranelift::builder()),
+        #[cfg(feature = "lightbeam")]
+        Strategy::Lightbeam => unimplemented!(),
+        #[cfg(not(feature = "lightbeam"))]
+        Strategy::Lightbeam => {
+            anyhow::bail!("lightbeam compilation strategy wasn't enabled at compile time");
+        }
+    }
+}
+
 fn round_up_to_pages(val: u64) -> u64 {
     let page_size = region::page::size() as u64;
     debug_assert!(page_size.is_power_of_two());
@@ -1270,12 +1356,35 @@ impl Default for Config {
     }
 }
 
+impl Clone for Config {
+    fn clone(&self) -> Config {
+        Config {
+            #[cfg(compiler)]
+            compiler: self.compiler.clone(),
+            tunables: self.tunables.clone(),
+            #[cfg(feature = "cache")]
+            cache_config: self.cache_config.clone(),
+            profiler: self.profiler.clone(),
+            features: self.features.clone(),
+            mem_creator: self.mem_creator.clone(),
+            allocation_strategy: self.allocation_strategy.clone(),
+            max_wasm_stack: self.max_wasm_stack,
+            wasm_backtrace_details_env_used: self.wasm_backtrace_details_env_used,
+            async_support: self.async_support,
+            #[cfg(feature = "async")]
+            async_stack_size: self.async_stack_size,
+            module_version: self.module_version.clone(),
+            parallel_compilation: self.parallel_compilation,
+            paged_memory_initialization: self.paged_memory_initialization,
+        }
+    }
+}
+
 impl fmt::Debug for Config {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.debug_struct("Config")
-            .field("debug_info", &self.tunables.generate_native_debuginfo)
+        let mut f = f.debug_struct("Config");
+        f.field("debug_info", &self.tunables.generate_native_debuginfo)
             .field("parse_wasm_debuginfo", &self.tunables.parse_wasm_debuginfo)
-            .field("strategy", &self.strategy)
             .field("wasm_threads", &self.features.threads)
             .field("wasm_reference_types", &self.features.reference_types)
             .field("wasm_bulk_memory", &self.features.bulk_memory)
@@ -1299,11 +1408,12 @@ impl fmt::Debug for Config {
                 "guard_before_linear_memory",
                 &self.tunables.guard_before_linear_memory,
             )
-            .field(
-                "flags",
-                &settings::Flags::new(self.flags.clone()).to_string(),
-            )
-            .finish()
+            .field("parallel_compilation", &self.parallel_compilation);
+        #[cfg(compiler)]
+        {
+            f.field("compiler", &self.compiler);
+        }
+        f.finish()
     }
 }
 
