@@ -1,10 +1,9 @@
-use crate::store::{StoreData, StoreInnermost, StoreOpaque, Stored};
+use crate::store::{StoreData, StoreOpaque, Stored};
 use crate::{
-    AsContext, AsContextMut, Engine, Extern, FuncType, Instance, InterruptHandle, StoreContext,
-    StoreContextMut, Trap, Val, ValType,
+    AsContext, AsContextMut, CallHook, Engine, Extern, FuncType, Instance, InterruptHandle,
+    StoreContext, StoreContextMut, Trap, Val, ValType,
 };
 use anyhow::{bail, Context as _, Result};
-use smallvec::{smallvec, SmallVec};
 use std::cmp::max;
 use std::error::Error;
 use std::fmt;
@@ -15,7 +14,7 @@ use std::pin::Pin;
 use std::ptr::NonNull;
 use std::sync::atomic::Ordering::Relaxed;
 use std::sync::Arc;
-use wasmtime_environ::wasm::{EntityIndex, FuncIndex};
+use wasmtime_environ::{EntityIndex, FuncIndex};
 use wasmtime_runtime::{
     raise_user_trap, ExportFunction, InstanceAllocator, InstanceHandle, OnDemandInstanceAllocator,
     VMCallerCheckedAnyfunc, VMContext, VMFunctionBody, VMFunctionImport, VMSharedSignatureIndex,
@@ -259,7 +258,7 @@ macro_rules! generate_wrap_async_func {
         {
             assert!(store.as_context().async_support(), concat!("cannot use `wrap", $num, "_async` without enabling async support on the config"));
             Func::wrap(store, move |mut caller: Caller<'_, T>, $($args: $args),*| {
-                let async_cx = caller.store.as_context_mut().opaque().async_cx();
+                let async_cx = caller.store.as_context_mut().0.async_cx();
                 let mut future = Pin::from(func(caller, $($args),*));
                 match unsafe { async_cx.block_on(future.as_mut()) } {
                     Ok(ret) => ret.into_fallible(),
@@ -300,18 +299,20 @@ impl Func {
     ///
     /// For more information about `Send + Sync + 'static` requirements on the
     /// `func`, see [`Func::wrap`](#why-send--sync--static).
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn new<T>(
         mut store: impl AsContextMut<Data = T>,
         ty: FuncType,
         func: impl Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<(), Trap> + Send + Sync + 'static,
     ) -> Self {
-        let mut store = store.as_context_mut().opaque();
+        let store = store.as_context_mut().0;
 
         // part of this unsafety is about matching the `T` to a `Store<T>`,
         // which is done through the `AsContextMut` bound above.
         unsafe {
             let host = HostFunc::new(store.engine(), ty, func);
-            host.into_func(&mut store)
+            host.into_func(store)
         }
     }
 
@@ -378,8 +379,8 @@ impl Func {
     /// # Ok(())
     /// # }
     /// ```
-    #[cfg(feature = "async")]
-    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
+    #[cfg(all(feature = "async", feature = "cranelift"))]
+    #[cfg_attr(nightlydoc, doc(cfg(all(feature = "async", feature = "cranelift"))))]
     pub fn new_async<T, F>(store: impl AsContextMut<Data = T>, ty: FuncType, func: F) -> Func
     where
         F: for<'a> Fn(
@@ -396,7 +397,7 @@ impl Func {
             "cannot use `new_async` without enabling async support in the config"
         );
         Func::new(store, ty, move |mut caller, params, results| {
-            let async_cx = caller.store.as_context_mut().opaque().async_cx();
+            let async_cx = caller.store.as_context_mut().0.async_cx();
             let mut future = Pin::from(func(caller, params, results));
             match unsafe { async_cx.block_on(future.as_mut()) } {
                 Ok(Ok(())) => Ok(()),
@@ -629,12 +630,12 @@ impl Func {
         mut store: impl AsContextMut<Data = T>,
         func: impl IntoFunc<T, Params, Results>,
     ) -> Func {
-        let mut store = store.as_context_mut().opaque();
+        let store = store.as_context_mut().0;
         // part of this unsafety is about matching the `T` to a `Store<T>`,
         // which is done through the `AsContextMut` bound above.
         unsafe {
             let host = HostFunc::wrap(store.engine(), func);
-            host.into_func(&mut store)
+            host.into_func(store)
         }
     }
 
@@ -734,7 +735,7 @@ impl Func {
         my_ty: FuncType,
         params: &[Val],
     ) -> Result<Box<[Val]>> {
-        let mut values_vec = write_params(&mut store.as_context_mut().opaque(), &my_ty, params)?;
+        let mut values_vec = write_params(store.0, &my_ty, params)?;
 
         // Call the trampoline.
         unsafe {
@@ -751,14 +752,10 @@ impl Func {
             })?;
         }
 
-        return Ok(read_results(
-            &mut store.as_context_mut().opaque(),
-            &my_ty,
-            &values_vec,
-        ));
+        return Ok(read_results(store.0, &my_ty, &values_vec));
 
         fn write_params(
-            store: &mut StoreOpaque<'_>,
+            store: &mut StoreOpaque,
             ty: &FuncType,
             params: &[Val],
         ) -> Result<Vec<u128>> {
@@ -777,6 +774,21 @@ impl Func {
 
             let mut values_vec = vec![0; max(params.len(), ty.results().len())];
 
+            // Whenever we pass `externref`s from host code to Wasm code, they
+            // go into the `VMExternRefActivationsTable`. But the table might be
+            // at capacity already, so check for that. If it is at capacity
+            // (unlikely) then do a GC to free up space. This is necessary
+            // because otherwise we would either keep filling up the bump chunk
+            // and making it larger and larger or we would always take the slow
+            // path when inserting references into the table.
+            if ty.as_wasm_func_type().externref_params_count()
+                > store
+                    .externref_activations_table()
+                    .bump_capacity_remaining()
+            {
+                store.gc();
+            }
+
             // Store the argument values into `values_vec`.
             let param_tys = ty.params();
             for ((arg, slot), ty) in params.iter().cloned().zip(&mut values_vec).zip(param_tys) {
@@ -791,18 +803,14 @@ impl Func {
                     bail!("cross-`Store` values are not currently supported");
                 }
                 unsafe {
-                    arg.write_value_to(store, slot);
+                    arg.write_value_without_gc(store, slot);
                 }
             }
 
             Ok(values_vec)
         }
 
-        fn read_results(
-            store: &mut StoreOpaque<'_>,
-            ty: &FuncType,
-            values_vec: &[u128],
-        ) -> Box<[Val]> {
+        fn read_results(store: &mut StoreOpaque, ty: &FuncType, values_vec: &[u128]) -> Box<[Val]> {
             let mut results = Vec::with_capacity(ty.results().len());
             for (index, ty) in ty.results().enumerate() {
                 unsafe {
@@ -817,7 +825,7 @@ impl Func {
     #[inline]
     pub(crate) fn caller_checked_anyfunc(
         &self,
-        store: &StoreInnermost,
+        store: &StoreOpaque,
     ) -> NonNull<VMCallerCheckedAnyfunc> {
         store.store_data()[self.0].export().anyfunc
     }
@@ -832,7 +840,7 @@ impl Func {
         Func(store.store_data_mut().insert(data))
     }
 
-    pub(crate) fn vmimport(&self, store: &mut StoreOpaque<'_>) -> VMFunctionImport {
+    pub(crate) fn vmimport(&self, store: &mut StoreOpaque) -> VMFunctionImport {
         unsafe {
             let f = self.caller_checked_anyfunc(store);
             VMFunctionImport {
@@ -852,48 +860,68 @@ impl Func {
         values_vec: *mut u128,
         func: &dyn Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<(), Trap>,
     ) -> Result<(), Trap> {
-        caller.store.0.entering_native_hook()?;
-        // We have a dynamic guarantee that `values_vec` has the right
-        // number of arguments and the right types of arguments. As a result
-        // we should be able to safely run through them all and read them.
-        const STACK_ARGS: usize = 4;
-        const STACK_RETURNS: usize = 2;
-        let mut args: SmallVec<[Val; STACK_ARGS]> = SmallVec::with_capacity(ty.params().len());
-        let mut store = caller.store.as_context_mut().opaque();
+        caller.store.0.call_hook(CallHook::CallingHost)?;
+
+        // Translate the raw JIT arguments in `values_vec` into a `Val` which
+        // we'll be passing as a slice. The storage for our slice-of-`Val` we'll
+        // be taking from the `Store`. We preserve our slice back into the
+        // `Store` after the hostcall, ideally amortizing the cost of allocating
+        // the storage across wasm->host calls.
+        //
+        // Note that we have a dynamic guarantee that `values_vec` is the
+        // appropriate length to both read all arguments from as well as store
+        // all results into.
+        let mut val_vec = caller.store.0.take_hostcall_val_storage();
+        debug_assert!(val_vec.is_empty());
+        let nparams = ty.params().len();
+        val_vec.reserve(nparams + ty.results().len());
         for (i, ty) in ty.params().enumerate() {
             unsafe {
-                let val = Val::read_value_from(&mut store, values_vec.add(i), ty);
-                args.push(val);
+                let val = Val::read_value_from(caller.store.0, values_vec.add(i), ty);
+                val_vec.push(val);
             }
         }
 
-        let mut returns: SmallVec<[Val; STACK_RETURNS]> =
-            smallvec![Val::null(); ty.results().len()];
+        val_vec.extend((0..ty.results().len()).map(|_| Val::null()));
+        let (params, results) = val_vec.split_at_mut(nparams);
+        func(caller.sub_caller(), params, results)?;
 
-        func(caller.sub_caller(), &args, &mut returns)?;
+        // See the comment in `Func::call_impl`'s `write_params` function.
+        if ty.as_wasm_func_type().externref_returns_count()
+            > caller
+                .store
+                .0
+                .externref_activations_table()
+                .bump_capacity_remaining()
+        {
+            caller.store.gc();
+        }
 
         // Unlike our arguments we need to dynamically check that the return
         // values produced are correct. There could be a bug in `func` that
         // produces the wrong number, wrong types, or wrong stores of
         // values, and we need to catch that here.
-        let mut store = caller.store.as_context_mut().opaque();
-        for (i, (ret, ty)) in returns.into_iter().zip(ty.results()).enumerate() {
+        for (i, (ret, ty)) in results.iter().zip(ty.results()).enumerate() {
             if ret.ty() != ty {
                 return Err(Trap::new(
                     "function attempted to return an incompatible value",
                 ));
             }
-            if !ret.comes_from_same_store(&store) {
+            if !ret.comes_from_same_store(caller.store.0) {
                 return Err(Trap::new(
                     "cross-`Store` values are not currently supported",
                 ));
             }
             unsafe {
-                ret.write_value_to(&mut store, values_vec.add(i));
+                ret.write_value_without_gc(caller.store.0, values_vec.add(i));
             }
         }
 
-        caller.store.0.exiting_native_hook()?;
+        // Restore our `val_vec` back into the store so it's usable for the next
+        // hostcall to reuse our own storage.
+        val_vec.truncate(0);
+        caller.store.0.save_hostcall_val_storage(val_vec);
+        caller.store.0.call_hook(CallHook::ReturningFromHost)?;
         Ok(())
     }
 
@@ -1042,7 +1070,7 @@ pub(crate) fn invoke_wasm_and_catch_traps<T>(
     unsafe {
         let exit = enter_wasm(store)?;
 
-        if let Err(trap) = store.0.exiting_native_hook() {
+        if let Err(trap) = store.0.call_hook(CallHook::CallingWasm) {
             exit_wasm(store, exit);
             return Err(trap);
         }
@@ -1053,8 +1081,8 @@ pub(crate) fn invoke_wasm_and_catch_traps<T>(
             closure,
         );
         exit_wasm(store, exit);
-        store.0.entering_native_hook()?;
-        result.map_err(Trap::from_runtime)
+        store.0.call_hook(CallHook::ReturningFromWasm)?;
+        result.map_err(Trap::from_runtime_box)
     }
 }
 
@@ -1135,7 +1163,7 @@ fn enter_wasm<T>(store: &mut StoreContextMut<'_, T>) -> Result<Option<usize>, Tr
             interrupts.stack_limit.store(usize::max_value(), Relaxed);
             return Err(Trap::new_wasm(
                 None,
-                wasmtime_environ::ir::TrapCode::Interrupt,
+                wasmtime_environ::TrapCode::Interrupt,
                 backtrace::Backtrace::new_unresolved(),
             ));
         }
@@ -1619,6 +1647,13 @@ impl<T> Caller<'_, T> {
         self.store.add_fuel(fuel)
     }
 
+    /// Synthetically consumes fuel from the store.
+    ///
+    /// For more information see [`Store::consume_fuel`](crate::Store::consume_fuel)
+    pub fn consume_fuel(&mut self, fuel: u64) -> Result<u64> {
+        self.store.consume_fuel(fuel)
+    }
+
     /// Configures this `Store` to trap whenever fuel runs out.
     ///
     /// For more information see
@@ -1739,16 +1774,15 @@ macro_rules! impl_into_func {
 
                         let ret = {
                             panic::catch_unwind(AssertUnwindSafe(|| {
-                                if let Err(trap) = caller.store.0.entering_native_hook() {
+                                if let Err(trap) = caller.store.0.call_hook(CallHook::CallingHost) {
                                     return R::fallible_from_trap(trap);
                                 }
-                                let mut _store = caller.sub_caller().store.opaque();
-                                $(let $args = $args::from_abi($args, &mut _store);)*
+                                $(let $args = $args::from_abi($args, caller.store.0);)*
                                 let r = func(
                                     caller.sub_caller(),
                                     $( $args, )*
                                 );
-                                if let Err(trap) = caller.store.0.exiting_native_hook() {
+                                if let Err(trap) = caller.store.0.call_hook(CallHook::ReturningFromHost) {
                                     return R::fallible_from_trap(trap);
                                 }
                                 r.into_fallible()
@@ -1768,11 +1802,10 @@ macro_rules! impl_into_func {
                                 // Because the wrapped function is not `unsafe`, we
                                 // can't assume it returned a value that is
                                 // compatible with this store.
-                                let mut store = caller.store.opaque();
-                                if !ret.compatible_with_store(&store) {
+                                if !ret.compatible_with_store(caller.store.0) {
                                     CallResult::Trap(cross_store_trap())
                                 } else {
-                                    match ret.into_abi_for_ret(&mut store, retptr) {
+                                    match ret.into_abi_for_ret(caller.store.0, retptr) {
                                         Ok(val) => CallResult::Ok(val),
                                         Err(trap) => CallResult::Trap(trap.into()),
                                     }
@@ -1883,6 +1916,7 @@ pub(crate) struct HostFunc {
 
 impl HostFunc {
     /// Analog of [`Func::new`]
+    #[cfg(compiler)]
     pub fn new<T>(
         engine: &Engine,
         ty: FuncType,
@@ -1891,11 +1925,11 @@ impl HostFunc {
         let ty_clone = ty.clone();
 
         // Create a trampoline that converts raw u128 values to `Val`
-        let func = Box::new(move |caller_vmctx, values_vec: *mut u128| unsafe {
+        let func = move |caller_vmctx, values_vec: *mut u128| unsafe {
             Caller::with(caller_vmctx, |caller| {
                 Func::invoke(caller, &ty_clone, values_vec, &func)
             })
-        });
+        };
 
         let (instance, trampoline) = crate::trampoline::create_function(&ty, func, engine)
             .expect("failed to create function");
@@ -1935,19 +1969,27 @@ impl HostFunc {
     ///
     /// Can only be inserted into stores with a matching `T` relative to when
     /// this `HostFunc` was first created.
-    pub unsafe fn to_func(self: &Arc<Self>, store: &mut StoreOpaque<'_>) -> Func {
+    pub unsafe fn to_func(self: &Arc<Self>, store: &mut StoreOpaque) -> Func {
         self.register_trampoline(store);
         let me = self.clone();
         Func(store.store_data_mut().insert(FuncData::SharedHost(me)))
     }
 
     /// Same as [`HostFunc::to_func`], different ownership.
-    unsafe fn into_func(self, store: &mut StoreOpaque<'_>) -> Func {
+    unsafe fn into_func(self, store: &mut StoreOpaque) -> Func {
         self.register_trampoline(store);
         Func(store.store_data_mut().insert(FuncData::Host(self)))
     }
 
-    unsafe fn register_trampoline(&self, store: &mut StoreOpaque<'_>) {
+    unsafe fn register_trampoline(&self, store: &mut StoreOpaque) {
+        // This assert is required to ensure that we can indeed safely insert
+        // `self` into the `store` provided, otherwise the type information we
+        // have listed won't be correct. This is possible to hit with the public
+        // API of Wasmtime, and should be documented in relevant functions.
+        assert!(
+            Engine::same(&self.engine, store.engine()),
+            "cannot use a store with a different engine than a linker was created with",
+        );
         let idx = self.export.anyfunc.as_ref().type_index;
         store.register_host_trampoline(idx, self.trampoline);
     }
