@@ -1,32 +1,68 @@
 //! Implements module serialization.
+//!
+//! This module implements the serialization format for `wasmtime::Module`.
+//! This includes both the binary format of the final artifact as well as
+//! validation on ingestion of artifacts.
+//!
+//! There are two main pieces of data associated with a binary artifact:
+//!
+//! 1. A list of compiled modules. The reason this is a list as opposed to one
+//!    singular module is that a module-linking module may encompass a number
+//!    of other modules.
+//! 2. Compilation metadata shared by all modules, including the global
+//!    `TypeTables` information. This metadata is validated for compilation
+//!    settings and also has information shared by all modules (such as the
+//!    shared `TypeTables`).
+//!
+//! Compiled modules are, at this time, represented as an ELF file. This ELF
+//! file contains all the necessary data needed to decode each individual
+//! module, and conveniently also handles things like alignment so we can
+//! actually directly `mmap` compilation artifacts from disk.
+//!
+//! With all this in mind, the current serialization format is as follows:
+//!
+//! * The first, primary, module starts the final artifact. This means that the
+//!   final artifact is actually, and conveniently, a valid ELF file. ELF files
+//!   don't place any restrictions on data coming after the ELF file itself,
+//!   so that's where everything else will go. Another reason for using this
+//!   format is that our compilation artifacts are then consumable by standard
+//!   debugging tools like `objdump` to poke around and see what's what.
+//!
+//! * Next, all other modules are encoded. Each module has its own alignment,
+//!   though, so modules aren't simply concatenated. Instead directly after an
+//!   ELF file there is a 64-bit little-endian integer which is the offset,
+//!   from the end of the previous ELF file, to the next ELF file.
+//!
+//! * Finally, once all modules have been encoded (there's always at least
+//!   one), the 8-byte value `u64::MAX` is encoded. Following this is a
+//!   number of fields:
+//!
+//!   1. The `HEADER` value
+//!   2. A byte indicating how long the next field is
+//!   3. A version string of the length of the previous byte value
+//!   4. A `bincode`-encoded `Metadata` structure.
+//!
+//!   This is hoped to help distinguish easily Wasmtime-based ELF files from
+//!   other random ELF files, as well as provide better error messages for
+//!   using wasmtime artifacts across versions.
+//!
+//! This format is implemented by the `to_bytes` and `from_mmap` function.
 
-use crate::{Engine, Module, OptLevel};
+use crate::{Engine, Module, ModuleVersionStrategy};
 use anyhow::{anyhow, bail, Context, Result};
-use bincode::Options;
+use object::read::elf::FileHeader;
+use object::{Bytes, File, Object, ObjectSection};
 use serde::{Deserialize, Serialize};
-use std::borrow::Cow;
-use std::fmt;
+use std::collections::BTreeMap;
+use std::convert::TryFrom;
+use std::path::Path;
 use std::str::FromStr;
 use std::sync::Arc;
-use std::{collections::HashMap, fmt::Display};
-use wasmtime_environ::{isa::TargetIsa, settings, Tunables};
-use wasmtime_jit::{
-    CompilationArtifacts, CompilationStrategy, CompiledModule, Compiler, TypeTables,
-};
+use wasmtime_environ::{Compiler, FlagValue, Tunables};
+use wasmtime_jit::{subslice_range, CompiledModule, CompiledModuleInfo, MmapVec, TypeTables};
 
 const HEADER: &[u8] = b"\0wasmtime-aot";
 
-fn bincode_options() -> impl Options {
-    // Use a variable-length integer encoding instead of fixed length. The
-    // module shown on #2318 gets compressed from ~160MB to ~110MB simply using
-    // this, presumably because there's a lot of 8-byte integers which generally
-    // have small values. Local testing shows that the deserialization
-    // performance, while higher, is in the few-percent range. For huge size
-    // savings this seems worthwhile to lose a small percentage of
-    // deserialization performance.
-    bincode::DefaultOptions::new().with_varint_encoding()
-}
-
 // This exists because `wasmparser::WasmFeatures` isn't serializable
 #[derive(Debug, Copy, Clone, Serialize, Deserialize)]
 struct WasmFeatures {
@@ -82,6 +118,12 @@ enum MyCow<'a, T> {
 }
 
 impl<'a, T> MyCow<'a, T> {
+    fn as_ref(&self) -> &T {
+        match self {
+            MyCow::Owned(val) => val,
+            MyCow::Borrowed(val) => val,
+        }
+    }
     fn unwrap_owned(self) -> T {
         match self {
             MyCow::Owned(val) => val,
@@ -111,16 +153,6 @@ impl<'a, 'b, T: Deserialize<'a>> Deserialize<'a> for MyCow<'b, T> {
     }
 }
 
-impl From<settings::OptLevel> for OptLevel {
-    fn from(level: settings::OptLevel) -> Self {
-        match level {
-            settings::OptLevel::Speed => OptLevel::Speed,
-            settings::OptLevel::SpeedAndSize => OptLevel::SpeedAndSize,
-            settings::OptLevel::None => OptLevel::None,
-        }
-    }
-}
-
 /// A small helper struct for serialized module upvars.
 #[derive(Serialize, Deserialize)]
 pub struct SerializedModuleUpvar {
@@ -163,58 +195,31 @@ impl SerializedModuleUpvar {
     }
 }
 
-#[derive(Serialize, Deserialize, Eq, PartialEq)]
-enum FlagValue {
-    Enum(Cow<'static, str>),
-    Num(u8),
-    Bool(bool),
-}
-
-impl From<settings::Value> for FlagValue {
-    fn from(v: settings::Value) -> Self {
-        match v.kind() {
-            settings::SettingKind::Enum => Self::Enum(v.as_enum().unwrap().into()),
-            settings::SettingKind::Num => Self::Num(v.as_num().unwrap()),
-            settings::SettingKind::Bool => Self::Bool(v.as_bool().unwrap()),
-            settings::SettingKind::Preset => unreachable!(),
-        }
-    }
-}
-
-impl Display for FlagValue {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match self {
-            Self::Enum(v) => v.fmt(f),
-            Self::Num(v) => v.fmt(f),
-            Self::Bool(v) => v.fmt(f),
-        }
-    }
+pub struct SerializedModule<'a> {
+    artifacts: Vec<MyCow<'a, MmapVec>>,
+    metadata: Metadata<'a>,
 }
 
 #[derive(Serialize, Deserialize)]
-pub struct SerializedModule<'a> {
+struct Metadata<'a> {
     target: String,
-    shared_flags: HashMap<String, FlagValue>,
-    isa_flags: HashMap<String, FlagValue>,
-    strategy: CompilationStrategy,
+    shared_flags: BTreeMap<String, FlagValue>,
+    isa_flags: BTreeMap<String, FlagValue>,
     tunables: Tunables,
     features: WasmFeatures,
-    artifacts: Vec<MyCow<'a, CompilationArtifacts>>,
     module_upvars: Vec<SerializedModuleUpvar>,
     types: MyCow<'a, TypeTables>,
 }
 
 impl<'a> SerializedModule<'a> {
+    #[cfg(compiler)]
     pub fn new(module: &'a Module) -> Self {
-        let compiler = module.engine().compiler();
         let artifacts = module
             .inner
             .artifact_upvars
             .iter()
-            .map(|m| MyCow::Borrowed(m.compilation_artifacts()))
-            .chain(Some(MyCow::Borrowed(
-                module.inner.module.compilation_artifacts(),
-            )))
+            .map(|m| MyCow::Borrowed(m.mmap()))
+            .chain(Some(MyCow::Borrowed(module.inner.module.mmap())))
             .collect::<Vec<_>>();
         let module_upvars = module
             .inner
@@ -224,153 +229,273 @@ impl<'a> SerializedModule<'a> {
             .collect::<Vec<_>>();
 
         Self::with_data(
-            compiler,
+            module.engine(),
             artifacts,
             module_upvars,
             MyCow::Borrowed(module.types()),
         )
     }
 
+    #[cfg(compiler)]
     pub fn from_artifacts(
-        compiler: &Compiler,
-        artifacts: &'a Vec<CompilationArtifacts>,
+        engine: &Engine,
+        artifacts: impl IntoIterator<Item = &'a MmapVec>,
         types: &'a TypeTables,
     ) -> Self {
         Self::with_data(
-            compiler,
-            artifacts.iter().map(MyCow::Borrowed).collect(),
+            engine,
+            artifacts.into_iter().map(MyCow::Borrowed).collect(),
             Vec::new(),
             MyCow::Borrowed(types),
         )
     }
 
+    #[cfg(compiler)]
     fn with_data(
-        compiler: &Compiler,
-        artifacts: Vec<MyCow<'a, CompilationArtifacts>>,
+        engine: &Engine,
+        artifacts: Vec<MyCow<'a, MmapVec>>,
         module_upvars: Vec<SerializedModuleUpvar>,
         types: MyCow<'a, TypeTables>,
     ) -> Self {
-        let isa = compiler.isa();
-
         Self {
-            target: isa.triple().to_string(),
-            shared_flags: isa
-                .flags()
-                .iter()
-                .map(|v| (v.name.to_owned(), v.into()))
-                .collect(),
-            isa_flags: isa
-                .isa_flags()
-                .into_iter()
-                .map(|v| (v.name.to_owned(), v.into()))
-                .collect(),
-            strategy: compiler.strategy(),
-            tunables: compiler.tunables().clone(),
-            features: compiler.features().into(),
             artifacts,
-            module_upvars,
-            types,
+            metadata: Metadata {
+                target: engine.compiler().triple().to_string(),
+                shared_flags: engine.compiler().flags(),
+                isa_flags: engine.compiler().isa_flags(),
+                tunables: engine.config().tunables.clone(),
+                features: (&engine.config().features).into(),
+                module_upvars,
+                types,
+            },
         }
     }
 
-    pub fn into_module(mut self, engine: &Engine) -> Result<Module> {
-        let compiler = engine.compiler();
-        let isa = compiler.isa();
+    pub fn into_module(self, engine: &Engine) -> Result<Module> {
+        let (main_module, modules, types, upvars) = self.into_parts(engine)?;
+        let modules = engine.run_maybe_parallel(modules, |(i, m)| {
+            CompiledModule::from_artifacts(i, m, &*engine.config().profiler)
+        })?;
 
-        self.check_triple(isa)?;
-        self.check_shared_flags(isa)?;
-        self.check_isa_flags(isa)?;
-        self.check_strategy(compiler)?;
-        self.check_tunables(compiler)?;
-        self.check_features(compiler)?;
+        Module::from_parts(engine, modules, main_module, Arc::new(types), &upvars)
+    }
 
-        let modules = CompiledModule::from_artifacts_list(
-            self.artifacts
-                .into_iter()
-                .map(|i| i.unwrap_owned())
-                .collect(),
-            engine.compiler().isa(),
-            &*engine.config().profiler,
-        )?;
+    pub fn into_parts(
+        mut self,
+        engine: &Engine,
+    ) -> Result<(
+        usize,
+        Vec<(MmapVec, Option<CompiledModuleInfo>)>,
+        TypeTables,
+        Vec<SerializedModuleUpvar>,
+    )> {
+        // Verify that the module we're loading matches the triple that `engine`
+        // is configured for. If compilation is disabled within engine then the
+        // assumed triple is the host itself.
+        #[cfg(compiler)]
+        let engine_triple = engine.compiler().triple();
+        #[cfg(not(compiler))]
+        let engine_triple = &target_lexicon::Triple::host();
+        self.check_triple(engine_triple)?;
 
-        assert!(!modules.is_empty());
+        // FIXME: Similar to `Module::from_binary` it should likely be validated
+        // here that when `cfg(not(compiler))` is true the isa/shared flags
+        // enabled for this precompiled module are compatible with the host
+        // itself, which `engine` is assumed to be running code for.
+        #[cfg(compiler)]
+        {
+            let compiler = engine.compiler();
+            self.check_shared_flags(compiler)?;
+            self.check_isa_flags(compiler)?;
+        }
+
+        self.check_tunables(&engine.config().tunables)?;
+        self.check_features(&engine.config().features)?;
+
+        assert!(!self.artifacts.is_empty());
+        let modules = self.artifacts.into_iter().map(|i| (i.unwrap_owned(), None));
 
         let main_module = modules.len() - 1;
 
-        Module::from_parts(
-            engine,
-            modules,
+        Ok((
             main_module,
-            Arc::new(self.types.unwrap_owned()),
-            &self.module_upvars,
-        )
+            modules.collect(),
+            self.metadata.types.unwrap_owned(),
+            self.metadata.module_upvars,
+        ))
     }
 
-    pub fn to_bytes(&self) -> Result<Vec<u8>> {
-        use std::io::Write;
+    pub fn to_bytes(&self, version_strat: &ModuleVersionStrategy) -> Result<Vec<u8>> {
+        // First up, create a linked-ish list of ELF files. For more
+        // information on this format, see the doc comment on this module.
+        // The only semi-tricky bit here is that we leave an
+        // offset-to-the-next-file between each set of ELF files. The list
+        // is then terminated with `u64::MAX`.
+        let mut ret = Vec::new();
+        for (i, obj) in self.artifacts.iter().enumerate() {
+            // Anything after the first object needs to respect the alignment of
+            // the object's sections, so insert padding as necessary. Note that
+            // the +8 to the length here is to accomodate the size we'll write
+            // to get to the next object.
+            if i > 0 {
+                let obj = File::parse(&obj.as_ref()[..])?;
+                let align = obj.sections().map(|s| s.align()).max().unwrap_or(0).max(1);
+                let align = usize::try_from(align).unwrap();
+                let new_size = align_to(ret.len() + 8, align);
+                ret.extend_from_slice(&(new_size as u64).to_le_bytes());
+                ret.resize(new_size, 0);
+            }
+            ret.extend_from_slice(obj.as_ref());
+        }
+        ret.extend_from_slice(&[0xff; 8]);
 
-        let mut bytes = Vec::new();
-
-        bytes.write_all(HEADER)?;
-
-        // Preface the data with a version so we can do a version check independent
-        // of the serialized data.
-        let version = env!("CARGO_PKG_VERSION");
+        // The last part of our artifact is the bincode-encoded `Metadata`
+        // section with a few other guards to help give better error messages.
+        ret.extend_from_slice(HEADER);
+        let version = match version_strat {
+            ModuleVersionStrategy::WasmtimeVersion => env!("CARGO_PKG_VERSION"),
+            ModuleVersionStrategy::Custom(c) => &c,
+            ModuleVersionStrategy::None => "",
+        };
+        // This precondition is checked in Config::module_version:
         assert!(
             version.len() < 256,
             "package version must be less than 256 bytes"
         );
-        bytes.write(&[version.len() as u8])?;
+        ret.push(version.len() as u8);
+        ret.extend_from_slice(version.as_bytes());
+        bincode::serialize_into(&mut ret, &self.metadata)?;
 
-        bytes.write_all(version.as_bytes())?;
-
-        bincode_options().serialize_into(&mut bytes, self)?;
-
-        Ok(bytes)
+        Ok(ret)
     }
 
-    pub fn from_bytes(bytes: &[u8], check_version: bool) -> Result<Self> {
-        if !bytes.starts_with(HEADER) {
-            bail!("bytes are not a compatible serialized wasmtime module");
-        }
+    pub fn from_bytes(bytes: &[u8], version_strat: &ModuleVersionStrategy) -> Result<Self> {
+        Self::from_mmap(MmapVec::from_slice(bytes)?, version_strat)
+    }
 
-        let bytes = &bytes[HEADER.len()..];
+    pub fn from_file(path: &Path, version_strat: &ModuleVersionStrategy) -> Result<Self> {
+        Self::from_mmap(
+            MmapVec::from_file(path).with_context(|| {
+                format!("failed to create file mapping for: {}", path.display())
+            })?,
+            version_strat,
+        )
+    }
 
-        if bytes.is_empty() {
+    pub fn from_mmap(mut mmap: MmapVec, version_strat: &ModuleVersionStrategy) -> Result<Self> {
+        // Artifacts always start with an ELF file, so read that first.
+        // Afterwards we continually read ELF files until we see the `u64::MAX`
+        // marker, meaning we've reached the end.
+        let first_module = read_file(&mut mmap)?;
+        let mut pos = first_module.len();
+        let mut artifacts = vec![MyCow::Owned(first_module)];
+
+        let metadata = loop {
+            if mmap.len() < 8 {
+                bail!("invalid serialized data");
+            }
+            let next_file_start = u64::from_le_bytes([
+                mmap[0], mmap[1], mmap[2], mmap[3], mmap[4], mmap[5], mmap[6], mmap[7],
+            ]);
+            if next_file_start == u64::MAX {
+                mmap.drain(..8);
+                break mmap;
+            }
+
+            // Remove padding leading up to the next file
+            let next_file_start = usize::try_from(next_file_start).unwrap();
+            let _padding = mmap.drain(..next_file_start - pos);
+            let data = read_file(&mut mmap)?;
+            pos = next_file_start + data.len();
+            artifacts.push(MyCow::Owned(data));
+        };
+
+        // Once we've reached the end we parse a `Metadata` object. This has a
+        // few guards up front which we process first, and eventually this
+        // bottoms out in a `bincode::deserialize` call.
+        let metadata = metadata
+            .strip_prefix(HEADER)
+            .ok_or_else(|| anyhow!("bytes are not a compatible serialized wasmtime module"))?;
+        if metadata.is_empty() {
             bail!("serialized data data is empty");
         }
-
-        let version_len = bytes[0] as usize;
-        if bytes.len() < version_len + 1 {
+        let version_len = metadata[0] as usize;
+        if metadata.len() < version_len + 1 {
             bail!("serialized data is malformed");
         }
 
-        if check_version {
-            let version = std::str::from_utf8(&bytes[1..1 + version_len])?;
-            if version != env!("CARGO_PKG_VERSION") {
-                bail!(
-                    "Module was compiled with incompatible Wasmtime version '{}'",
-                    version
-                );
+        match version_strat {
+            ModuleVersionStrategy::WasmtimeVersion => {
+                let version = std::str::from_utf8(&metadata[1..1 + version_len])?;
+                if version != env!("CARGO_PKG_VERSION") {
+                    bail!(
+                        "Module was compiled with incompatible Wasmtime version '{}'",
+                        version
+                    );
+                }
             }
+            ModuleVersionStrategy::Custom(v) => {
+                let version = std::str::from_utf8(&metadata[1..1 + version_len])?;
+                if version != v {
+                    bail!(
+                        "Module was compiled with incompatible version '{}'",
+                        version
+                    );
+                }
+            }
+            ModuleVersionStrategy::None => { /* ignore the version info, accept all */ }
         }
 
-        Ok(bincode_options()
-            .deserialize::<SerializedModule<'_>>(&bytes[1 + version_len..])
-            .context("deserialize compilation artifacts")?)
+        let metadata = bincode::deserialize::<Metadata>(&metadata[1 + version_len..])
+            .context("deserialize compilation artifacts")?;
+
+        return Ok(SerializedModule {
+            artifacts,
+            metadata,
+        });
+
+        /// This function will drain the beginning contents of `mmap` which
+        /// correspond to an ELF object file. The ELF file is only very lightly
+        /// validated.
+        ///
+        /// The `mmap` passed in will be reset to just after the ELF file, and
+        /// the `MmapVec` returned represents the extend of the ELF file
+        /// itself.
+        fn read_file(mmap: &mut MmapVec) -> Result<MmapVec> {
+            use object::NativeEndian as NE;
+            // There's not actually a great utility for figuring out where
+            // the end of an ELF file is in the `object` crate. In lieu of that
+            // we build our own which leverages the format of ELF files, which
+            // is that the header comes first, that tells us where the section
+            // headers are, and for our ELF files the end of the file is the
+            // end of the section headers.
+            let mut bytes = Bytes(mmap);
+            let header = bytes
+                .read::<object::elf::FileHeader64<NE>>()
+                .map_err(|()| anyhow!("artifact truncated, can't read header"))?;
+            if !header.is_supported() {
+                bail!("invalid elf header");
+            }
+            let sections = header
+                .section_headers(NE, &mmap[..])
+                .context("failed to read section headers")?;
+            let range = subslice_range(object::bytes_of_slice(sections), mmap);
+            Ok(mmap.drain(..range.end))
+        }
     }
 
-    fn check_triple(&self, isa: &dyn TargetIsa) -> Result<()> {
-        let triple = target_lexicon::Triple::from_str(&self.target).map_err(|e| anyhow!(e))?;
+    fn check_triple(&self, other: &target_lexicon::Triple) -> Result<()> {
+        let triple =
+            target_lexicon::Triple::from_str(&self.metadata.target).map_err(|e| anyhow!(e))?;
 
-        if triple.architecture != isa.triple().architecture {
+        if triple.architecture != other.architecture {
             bail!(
                 "Module was compiled for architecture '{}'",
                 triple.architecture
             );
         }
 
-        if triple.operating_system != isa.triple().operating_system {
+        if triple.operating_system != other.operating_system {
             bail!(
                 "Module was compiled for operating system '{}'",
                 triple.operating_system
@@ -380,13 +505,11 @@ impl<'a> SerializedModule<'a> {
         Ok(())
     }
 
-    fn check_shared_flags(&mut self, isa: &dyn TargetIsa) -> Result<()> {
-        let mut shared_flags = std::mem::take(&mut self.shared_flags);
-        for value in isa.flags().iter() {
-            let name = value.name;
-            match shared_flags.remove(name) {
+    fn check_shared_flags(&mut self, compiler: &dyn Compiler) -> Result<()> {
+        let mut shared_flags = std::mem::take(&mut self.metadata.shared_flags);
+        for (name, host) in compiler.flags() {
+            match shared_flags.remove(&name) {
                 Some(v) => {
-                    let host: FlagValue = value.into();
                     if v != host {
                         bail!("Module was compiled with a different '{}' setting: expected '{}' but host has '{}'", name, v, host);
                     }
@@ -405,12 +528,10 @@ impl<'a> SerializedModule<'a> {
         Ok(())
     }
 
-    fn check_isa_flags(&mut self, isa: &dyn TargetIsa) -> Result<()> {
-        let mut isa_flags = std::mem::take(&mut self.isa_flags);
-        for value in isa.isa_flags().into_iter() {
-            let name = value.name;
-            let host: FlagValue = value.into();
-            match isa_flags.remove(name) {
+    fn check_isa_flags(&mut self, compiler: &dyn Compiler) -> Result<()> {
+        let mut isa_flags = std::mem::take(&mut self.metadata.isa_flags);
+        for (name, host) in compiler.isa_flags() {
+            match isa_flags.remove(&name) {
                 Some(v) => match (&v, &host) {
                     (FlagValue::Bool(v), FlagValue::Bool(host)) => {
                         // ISA flags represent CPU features; for boolean values, only
@@ -440,25 +561,6 @@ impl<'a> SerializedModule<'a> {
         Ok(())
     }
 
-    fn check_strategy(&self, compiler: &Compiler) -> Result<()> {
-        #[allow(unreachable_patterns)]
-        let matches = match (self.strategy, compiler.strategy()) {
-            (CompilationStrategy::Auto, CompilationStrategy::Auto)
-            | (CompilationStrategy::Auto, CompilationStrategy::Cranelift)
-            | (CompilationStrategy::Cranelift, CompilationStrategy::Auto)
-            | (CompilationStrategy::Cranelift, CompilationStrategy::Cranelift) => true,
-            #[cfg(feature = "lightbeam")]
-            (CompilationStrategy::Lightbeam, CompilationStrategy::Lightbeam) => true,
-            _ => false,
-        };
-
-        if !matches {
-            bail!("Module was compiled with strategy '{:?}'", self.strategy);
-        }
-
-        Ok(())
-    }
-
     fn check_int<T: Eq + std::fmt::Display>(found: T, expected: T, feature: &str) -> Result<()> {
         if found == expected {
             return Ok(());
@@ -485,7 +587,7 @@ impl<'a> SerializedModule<'a> {
         );
     }
 
-    fn check_tunables(&self, compiler: &Compiler) -> Result<()> {
+    fn check_tunables(&mut self, other: &Tunables) -> Result<()> {
         let Tunables {
             static_memory_bound,
             static_memory_offset_guard_size,
@@ -496,9 +598,10 @@ impl<'a> SerializedModule<'a> {
             consume_fuel,
             static_memory_bound_is_maximum,
             guard_before_linear_memory,
-        } = self.tunables;
 
-        let other = compiler.tunables();
+            // This doesn't affect compilation, it's just a runtime setting.
+            dynamic_memory_growth_reserve: _,
+        } = self.metadata.tunables;
 
         Self::check_int(
             static_memory_bound,
@@ -541,7 +644,7 @@ impl<'a> SerializedModule<'a> {
         Ok(())
     }
 
-    fn check_features(&self, compiler: &Compiler) -> Result<()> {
+    fn check_features(&mut self, other: &wasmparser::WasmFeatures) -> Result<()> {
         let WasmFeatures {
             reference_types,
             multi_value,
@@ -554,9 +657,8 @@ impl<'a> SerializedModule<'a> {
             multi_memory,
             exceptions,
             memory64,
-        } = self.features;
+        } = self.metadata.features;
 
-        let other = compiler.features();
         Self::check_bool(
             reference_types,
             other.reference_types,
@@ -605,10 +707,17 @@ impl<'a> SerializedModule<'a> {
     }
 }
 
+/// Aligns the `val` specified up to `align`, which must be a power of two
+fn align_to(val: usize, align: usize) -> usize {
+    debug_assert!(align.is_power_of_two());
+    (val + (align - 1)) & (!(align - 1))
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
     use crate::Config;
+    use std::borrow::Cow;
 
     #[test]
     fn test_architecture_mismatch() -> Result<()> {
@@ -616,7 +725,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.target = "unknown-generic-linux".to_string();
+        serialized.metadata.target = "unknown-generic-linux".to_string();
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
@@ -635,7 +744,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.target = format!(
+        serialized.metadata.target = format!(
             "{}-generic-unknown",
             target_lexicon::Triple::host().architecture
         );
@@ -657,7 +766,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.shared_flags.insert(
+        serialized.metadata.shared_flags.insert(
             "opt_level".to_string(),
             FlagValue::Enum(Cow::Borrowed("none")),
         );
@@ -681,6 +790,7 @@ mod test {
         let mut serialized = SerializedModule::new(&module);
 
         serialized
+            .metadata
             .isa_flags
             .insert("not_a_flag".to_string(), FlagValue::Bool(true));
 
@@ -702,7 +812,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.strategy = CompilationStrategy::Lightbeam;
+        serialized.metadata.strategy = CompilationStrategy::Lightbeam;
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
@@ -721,7 +831,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.tunables.static_memory_offset_guard_size = 0;
+        serialized.metadata.tunables.static_memory_offset_guard_size = 0;
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
@@ -740,7 +850,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.tunables.interruptable = false;
+        serialized.metadata.tunables.interruptable = false;
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
@@ -757,7 +867,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.tunables.interruptable = true;
+        serialized.metadata.tunables.interruptable = true;
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
@@ -779,7 +889,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.features.simd = false;
+        serialized.metadata.features.simd = false;
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
@@ -793,7 +903,7 @@ mod test {
         let module = Module::new(&engine, "(module)")?;
 
         let mut serialized = SerializedModule::new(&module);
-        serialized.features.simd = true;
+        serialized.metadata.features.simd = true;
 
         match serialized.into_module(&engine) {
             Ok(_) => unreachable!(),
