@@ -1,4 +1,82 @@
-use crate::{module::ModuleRegistry, Engine, Module, Trap};
+//! Wasmtime's "store" type
+//!
+//! This module, and its submodules, contain the `Store` type and various types
+//! used to interact with it. At first glance this is a pretty confusing module
+//! where you need to know the difference between:
+//!
+//! * `Store<T>`
+//! * `StoreContext<T>`
+//! * `StoreContextMut<T>`
+//! * `AsContext`
+//! * `AsContextMut`
+//! * `StoreInner<T>`
+//! * `StoreOpaque`
+//! * `StoreData`
+//!
+//! There's... quite a lot going on here, and it's easy to be confused. This
+//! comment is ideally going to serve the purpose of clarifying what all these
+//! types are for and why they're motivated.
+//!
+//! First it's important to know what's "internal" and what's "external". Almost
+//! everything above is defined as `pub`, but only some of the items are
+//! reexported to the outside world to be usable from this crate. Otherwise all
+//! items are `pub` within this `store` module, and the `store` module is
+//! private to the `wasmtime` crate. Notably `Store<T>`, `StoreContext<T>`,
+//! `StoreContextMut<T>`, `AsContext`, and `AsContextMut` are all public
+//! interfaces to the `wasmtime` crate. You can think of these as:
+//!
+//! * `Store<T>` - an owned reference to a store, the "root of everything"
+//! * `StoreContext<T>` - basically `&StoreInner<T>`
+//! * `StoreContextMut<T>` - more-or-less `&mut StoreInner<T>` with caveats.
+//!   Explained later.
+//! * `AsContext` - similar to `AsRef`, but produces `StoreContext<T>`
+//! * `AsContextMut` - similar to `AsMut`, but produces `StoreContextMut<T>`
+//!
+//! Next comes the internal structure of the `Store<T>` itself. This looks like:
+//!
+//! * `Store<T>` - this type is just a pointer large. It's primarily just
+//!   intended to be consumed by the outside world. Note that the "just a
+//!   pointer large" is a load-bearing implementation detail in Wasmtime. This
+//!   enables it to store a pointer to its own trait object which doesn't need
+//!   to change over time.
+//!
+//! * `StoreInner<T>` - the first layer of the contents of a `Store<T>`, what's
+//!   stored inside the `Box`. This is the general Rust pattern when one struct
+//!   is a layer over another. The surprising part, though, is that this is
+//!   further subdivided. This structure only contains things which actually
+//!   need `T` itself. The downside of this structure is that it's always
+//!   generic and means that code is monomorphized into consumer crates. We
+//!   strive to have things be as monomorphic as possible in `wasmtime` so this
+//!   type is not heavily used.
+//!
+//! * `StoreOpaque` - this is the primary contents of the `StoreInner<T>` type.
+//!   Stored inline in the outer type the "opaque" here means that it's a
+//!   "store" but it doesn't have access to the `T`. This is the primary
+//!   "internal" reference that Wasmtime uses since `T` is rarely needed by the
+//!   internals of Wasmtime.
+//!
+//! * `StoreData` - this is a final helper struct stored within `StoreOpaque`.
+//!   All references of Wasm items into a `Store` are actually indices into a
+//!   table in this structure, and the `StoreData` being separate makes it a bit
+//!   easier to manage/define/work with. There's no real fundamental reason this
+//!   is split out, although sometimes it's useful to have separate borrows into
+//!   these tables than the `StoreOpaque`.
+//!
+//! A major caveat with these representations is that the internal `&mut
+//! StoreInner<T>` is never handed out publicly to consumers of this crate, only
+//! through a wrapper of `StoreContextMut<'_, T>`. The reason for this is that
+//! we want to provide mutable, but not destructive, access to the contents of a
+//! `Store`. For example if a `StoreInner<T>` were replaced with some other
+//! `StoreInner<T>` then that would drop live instances, possibly those
+//! currently executing beneath the current stack frame. This would not be a
+//! safe operation.
+//!
+//! This means, though, that the `wasmtime` crate, which liberally uses `&mut
+//! StoreOpaque` internally, has to be careful to never actually destroy the
+//! contents of `StoreOpaque`. This is an invariant that we, as the authors of
+//! `wasmtime`, must uphold for the public interface to be safe.
+
+use crate::{module::ModuleRegistry, Engine, Module, Trap, Val};
 use anyhow::{bail, Result};
 use std::cell::UnsafeCell;
 use std::collections::HashMap;
@@ -7,7 +85,7 @@ use std::error::Error;
 use std::fmt;
 use std::future::Future;
 use std::marker;
-use std::mem::ManuallyDrop;
+use std::mem::{self, ManuallyDrop};
 use std::ops::{Deref, DerefMut};
 use std::pin::Pin;
 use std::ptr;
@@ -79,14 +157,80 @@ pub struct Store<T> {
     inner: ManuallyDrop<Box<StoreInner<T>>>,
 }
 
+#[derive(Copy, Clone, Debug)]
+/// Passed to the argument of [`Store::call_hook`] to indicate a state transition in
+/// the WebAssembly VM.
+pub enum CallHook {
+    /// Indicates the VM is calling a WebAssembly function, from the host.
+    CallingWasm,
+    /// Indicates the VM is returning from a WebAssembly function, to the host.
+    ReturningFromWasm,
+    /// Indicates the VM is calling a host function, from WebAssembly.
+    CallingHost,
+    /// Indicates the VM is returning from a host function, to WebAssembly.
+    ReturningFromHost,
+}
+
+impl CallHook {
+    /// Indicates the VM is entering host code (exiting WebAssembly code)
+    pub fn entering_host(&self) -> bool {
+        match self {
+            CallHook::ReturningFromWasm | CallHook::CallingHost => true,
+            _ => false,
+        }
+    }
+    /// Indicates the VM is exiting host code (entering WebAssembly code)
+    pub fn exiting_host(&self) -> bool {
+        match self {
+            CallHook::ReturningFromHost | CallHook::CallingWasm => true,
+            _ => false,
+        }
+    }
+}
+
+/// Internal contents of a `Store<T>` that live on the heap.
+///
+/// The members of this struct are those that need to be generic over `T`, the
+/// store's internal type storage. Otherwise all things that don't rely on `T`
+/// should go into `StoreOpaque`.
 pub struct StoreInner<T> {
-    // This `StoreInner<T>` structure has references to itself. These aren't
+    /// Generic metadata about the store that doesn't need access to `T`.
+    inner: StoreOpaque,
+
+    limiter: Option<Box<dyn FnMut(&mut T) -> &mut (dyn crate::ResourceLimiter) + Send + Sync>>,
+    call_hook: Option<Box<dyn FnMut(&mut T, CallHook) -> Result<(), crate::Trap> + Send + Sync>>,
+    // for comments about `ManuallyDrop`, see `Store::into_data`
+    data: ManuallyDrop<T>,
+}
+
+// Forward methods on `StoreOpaque` to also being on `StoreInner<T>`
+impl<T> Deref for StoreInner<T> {
+    type Target = StoreOpaque;
+    fn deref(&self) -> &Self::Target {
+        &self.inner
+    }
+}
+
+impl<T> DerefMut for StoreInner<T> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.inner
+    }
+}
+
+/// Monomorphic storage for a `Store<T>`.
+///
+/// This structure contains the bulk of the metadata about a `Store`. This is
+/// used internally in Wasmtime when dependence on the `T` of `Store<T>` isn't
+/// necessary, allowing code to be monomorphic and compiled into the `wasmtime`
+/// crate itself.
+pub struct StoreOpaque {
+    // This `StoreOpaque` structure has references to itself. These aren't
     // immediately evident, however, so we need to tell the compiler that it
     // contains self-references. This notably suppresses `noalias` annotations
     // when this shows up in compiled code because types of this structure do
-    // indeed alias itself. The best example of this is `StoreOpaque` which
-    // contains a `&mut StoreInner` and a `*mut dyn Store` which are actually
-    // the same pointer, indeed aliasing!
+    // indeed alias itself. An example of this is `default_callee` holds a
+    // `*mut dyn Store` to the address of this `StoreOpaque` itself, indeed
+    // aliasing!
     //
     // It's somewhat unclear to me at this time if this is 100% sufficient to
     // get all the right codegen in all the right places. For example does
@@ -102,30 +246,7 @@ pub struct StoreInner<T> {
     // least telling the compiler something about all the aliasing happening
     // within a `Store`.
     _marker: marker::PhantomPinned,
-    inner: StoreInnermost,
-    limiter: Option<Box<dyn FnMut(&mut T) -> &mut (dyn crate::ResourceLimiter) + Send + Sync>>,
-    entering_native_hook: Option<Box<dyn FnMut(&mut T) -> Result<(), crate::Trap> + Send + Sync>>,
-    exiting_native_hook: Option<Box<dyn FnMut(&mut T) -> Result<(), crate::Trap> + Send + Sync>>,
-    // for comments about `ManuallyDrop`, see `Store::into_data`
-    data: ManuallyDrop<T>,
-}
 
-impl<T> Deref for StoreInner<T> {
-    type Target = StoreInnermost;
-    fn deref(&self) -> &Self::Target {
-        &self.inner
-    }
-}
-impl<T> DerefMut for StoreInner<T> {
-    fn deref_mut(&mut self) -> &mut Self::Target {
-        &mut self.inner
-    }
-}
-
-// I apologize for the convoluted structure and the terrible naming of this struct.
-// This exists so that most of wasmtime can be monomorphic on StoreInnermost, without
-// having to care about the generic in StoreInner<T>.
-pub struct StoreInnermost {
     engine: Engine,
     interrupts: Arc<VMInterrupts>,
     instances: Vec<StoreInstance>,
@@ -148,6 +269,11 @@ pub struct StoreInnermost {
     out_of_gas_behavior: OutOfGas,
     store_data: StoreData,
     default_callee: InstanceHandle,
+
+    /// Used to optimzed wasm->host calls when the host function is defined with
+    /// `Func::new` to avoid allocating a new vector each time a function is
+    /// called.
+    hostcall_val_storage: Vec<Val>,
 }
 
 #[cfg(feature = "async")]
@@ -164,6 +290,68 @@ unsafe impl Send for AsyncState {}
 #[cfg(feature = "async")]
 unsafe impl Sync for AsyncState {}
 
+/// An RAII type to automatically mark a region of code as unsafe for GC.
+pub(crate) struct AutoAssertNoGc<T>
+where
+    T: std::ops::DerefMut<Target = StoreOpaque>,
+{
+    #[cfg(debug_assertions)]
+    prev_okay: bool,
+    store: T,
+}
+
+impl<T> AutoAssertNoGc<T>
+where
+    T: std::ops::DerefMut<Target = StoreOpaque>,
+{
+    pub fn new(mut store: T) -> Self {
+        drop(&mut store);
+        #[cfg(debug_assertions)]
+        {
+            let prev_okay = store.externref_activations_table.set_gc_okay(false);
+            return AutoAssertNoGc { store, prev_okay };
+        }
+        #[cfg(not(debug_assertions))]
+        {
+            return AutoAssertNoGc { store };
+        }
+    }
+}
+
+impl<T> std::ops::Deref for AutoAssertNoGc<T>
+where
+    T: std::ops::DerefMut<Target = StoreOpaque>,
+{
+    type Target = T;
+
+    fn deref(&self) -> &Self::Target {
+        &self.store
+    }
+}
+
+impl<T> std::ops::DerefMut for AutoAssertNoGc<T>
+where
+    T: std::ops::DerefMut<Target = StoreOpaque>,
+{
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.store
+    }
+}
+
+impl<T> Drop for AutoAssertNoGc<T>
+where
+    T: std::ops::DerefMut<Target = StoreOpaque>,
+{
+    fn drop(&mut self) {
+        #[cfg(debug_assertions)]
+        {
+            self.store
+                .externref_activations_table
+                .set_gc_okay(self.prev_okay);
+        }
+    }
+}
+
 /// Used to associate instances with the store.
 ///
 /// This is needed to track if the instance was allocated explicitly with the on-demand
@@ -194,7 +382,7 @@ impl<T> Store<T> {
     /// tables created to 10,000. This can be overridden with the
     /// [`Store::limiter`] configuration method.
     pub fn new(engine: &Engine, data: T) -> Self {
-        let finished_functions = &Default::default();
+        let functions = &Default::default();
         // Wasmtime uses the callee argument to host functions to learn about
         // the original pointer to the `Store` itself, allowing it to
         // reconstruct a `StoreContextMut<T>`. When we initially call a `Func`,
@@ -206,17 +394,19 @@ impl<T> Store<T> {
             OnDemandInstanceAllocator::default()
                 .allocate(InstanceAllocationRequest {
                     host_state: Box::new(()),
-                    finished_functions,
+                    image_base: 0,
+                    functions,
                     shared_signatures: None.into(),
                     imports: Default::default(),
                     module: Arc::new(wasmtime_environ::Module::default()),
                     store: None,
+                    wasm_data: &[],
                 })
                 .expect("failed to allocate default callee")
         };
         let mut inner = Box::new(StoreInner {
-            _marker: marker::PhantomPinned,
-            inner: StoreInnermost {
+            inner: StoreOpaque {
+                _marker: marker::PhantomPinned,
                 engine: engine.clone(),
                 interrupts: Default::default(),
                 instances: Vec::new(),
@@ -239,19 +429,26 @@ impl<T> Store<T> {
                 out_of_gas_behavior: OutOfGas::Trap,
                 store_data: StoreData::new(),
                 default_callee,
+                hostcall_val_storage: Vec::new(),
             },
             limiter: None,
-            entering_native_hook: None,
-            exiting_native_hook: None,
+            call_hook: None,
             data: ManuallyDrop::new(data),
         });
 
         // Once we've actually allocated the store itself we can configure the
-        // trait object pointer of the default callee.
-        let store = StoreContextMut(&mut *inner).opaque().traitobj;
+        // trait object pointer of the default callee. Note the erasure of the
+        // lifetime here into `'static`, so in general usage of this trait
+        // object must be strictly bounded to the `Store` itself, and is a
+        // variant that we have to maintain throughout Wasmtime.
         unsafe {
-            inner.default_callee.set_store(store);
+            let traitobj = std::mem::transmute::<
+                *mut (dyn wasmtime_runtime::Store + '_),
+                *mut (dyn wasmtime_runtime::Store + 'static),
+            >(&mut *inner);
+            inner.default_callee.set_store(traitobj);
         }
+
         Self {
             inner: ManuallyDrop::new(inner),
         }
@@ -327,56 +524,25 @@ impl<T> Store<T> {
         inner.limiter = Some(Box::new(limiter));
     }
 
-    /// Configure a function that runs each time the host resumes execution from
-    /// WebAssembly.
+    /// Configure a function that runs on calls and returns between WebAssembly
+    /// and host code.
     ///
-    /// This hook is called in two circumstances:
-    ///
-    /// * When WebAssembly calls a function defined by the host, this hook is
-    ///   called before other host code runs.
-    /// * When WebAssembly returns back to the host after being called, this
-    ///   hook is called.
-    ///
-    /// This method can be used with [`Store::exiting_native_code_hook`] to track
-    /// execution time of WebAssembly, for example, by starting/stopping timers
-    /// in the enter/exit hooks.
+    /// The function is passed a [`CallHook`] argument, which indicates which
+    /// state transition the VM is making.
     ///
     /// This function may return a [`Trap`]. If a trap is returned when an
     /// import was called, it is immediately raised as-if the host import had
     /// returned the trap. If a trap is returned after wasm returns to the host
     /// then the wasm function's result is ignored and this trap is returned
     /// instead.
-    pub fn entering_native_code_hook(
+    ///
+    /// After this function returns a trap, it may be called for subsequent returns
+    /// to host or wasm code as the trap propogates to the root call.
+    pub fn call_hook(
         &mut self,
-        hook: impl FnMut(&mut T) -> Result<(), Trap> + Send + Sync + 'static,
+        hook: impl FnMut(&mut T, CallHook) -> Result<(), Trap> + Send + Sync + 'static,
     ) {
-        self.inner.entering_native_hook = Some(Box::new(hook));
-    }
-
-    /// Configure a function that runs just before WebAssembly code starts
-    /// executing.
-    ///
-    /// The closure provided is called in two circumstances:
-    ///
-    /// * When the host calls a WebAssembly function, the hook is called just
-    ///   before WebAssembly starts executing.
-    /// * When a host function returns back to WebAssembly this hook is called
-    ///   just before the return.
-    ///
-    /// This method can be used with [`Store::entering_native_code_hook`] to track
-    /// execution time of WebAssembly, for example, by starting/stopping timers
-    /// in the enter/exit hooks.
-    ///
-    /// This function may return a [`Trap`]. If a trap is returned when an
-    /// imported host function is returning, then the imported host function's
-    /// result is ignored and the trap is raised. If a trap is returned when
-    /// the host is about to start executing WebAssembly, then no WebAssembly
-    /// code is run and the trap is returned instead.
-    pub fn exiting_native_code_hook(
-        &mut self,
-        hook: impl FnMut(&mut T) -> Result<(), Trap> + Send + Sync + 'static,
-    ) {
-        self.inner.exiting_native_hook = Some(Box::new(hook));
+        self.inner.call_hook = Some(Box::new(hook));
     }
 
     /// Returns the [`Engine`] that this store is associated with.
@@ -510,6 +676,29 @@ impl<T> Store<T> {
         self.inner.add_fuel(fuel)
     }
 
+    /// Synthetically consumes fuel from this [`Store`].
+    ///
+    /// For this method to work fuel consumption must be enabled via
+    /// [`Config::consume_fuel`](crate::Config::consume_fuel).
+    ///
+    /// WebAssembly execution will automatically consume fuel but if so desired
+    /// the embedder can also consume fuel manually to account for relative
+    /// costs of host functions, for example.
+    ///
+    /// This function will attempt to consume `fuel` units of fuel from within
+    /// this store. If the remaining amount of fuel allows this then `Ok(N)` is
+    /// returned where `N` is the amount of remaining fuel. Otherwise an error
+    /// is returned and no fuel is consumed.
+    ///
+    /// # Errors
+    ///
+    /// This function will return an either either if fuel consumption via
+    /// [`Config`](crate::Config) is disabled or if `fuel` exceeds the amount
+    /// of remaining fuel within this store.
+    pub fn consume_fuel(&mut self, fuel: u64) -> Result<u64> {
+        self.inner.consume_fuel(fuel)
+    }
+
     /// Configures a [`Store`] to generate a [`Trap`] whenever it runs out of
     /// fuel.
     ///
@@ -644,6 +833,13 @@ impl<'a, T> StoreContextMut<'a, T> {
         self.0.add_fuel(fuel)
     }
 
+    /// Synthetically consume fuel from this store.
+    ///
+    /// For more information see [`Store::consume_fuel`]
+    pub fn consume_fuel(&mut self, fuel: u64) -> Result<u64> {
+        self.0.consume_fuel(fuel)
+    }
+
     /// Configures this `Store` to trap whenever fuel runs out.
     ///
     /// For more information see [`Store::out_of_fuel_trap`]
@@ -677,24 +873,16 @@ impl<T> StoreInner<T> {
         Some(accessor(&mut self.data))
     }
 
-    pub fn entering_native_hook(&mut self) -> Result<(), Trap> {
-        if let Some(hook) = &mut self.entering_native_hook {
-            hook(&mut self.data)
-        } else {
-            Ok(())
-        }
-    }
-
-    pub fn exiting_native_hook(&mut self) -> Result<(), Trap> {
-        if let Some(hook) = &mut self.exiting_native_hook {
-            hook(&mut self.data)
+    pub fn call_hook(&mut self, s: CallHook) -> Result<(), Trap> {
+        if let Some(hook) = &mut self.call_hook {
+            hook(&mut self.data, s)
         } else {
             Ok(())
         }
     }
 }
 
-impl StoreInnermost {
+impl StoreOpaque {
     pub fn bump_resource_counts(&mut self, module: &Module) -> Result<()> {
         fn bump(slot: &mut usize, max: usize, amt: usize, desc: &str) -> Result<()> {
             let new = slot.saturating_add(amt);
@@ -724,6 +912,7 @@ impl StoreInnermost {
 
         Ok(())
     }
+
     #[inline]
     pub fn async_support(&self) -> bool {
         cfg!(feature = "async") && self.engine().config().async_support
@@ -734,10 +923,12 @@ impl StoreInnermost {
         &self.engine
     }
 
+    #[inline]
     pub fn store_data(&self) -> &StoreData {
         &self.store_data
     }
 
+    #[inline]
     pub fn store_data_mut(&mut self) -> &mut StoreData {
         &mut self.store_data
     }
@@ -929,6 +1120,20 @@ impl StoreInnermost {
         Ok(())
     }
 
+    fn consume_fuel(&mut self, fuel: u64) -> Result<u64> {
+        let consumed_ptr = unsafe { &mut *self.interrupts.fuel_consumed.get() };
+        match i64::try_from(fuel)
+            .ok()
+            .and_then(|fuel| consumed_ptr.checked_add(fuel))
+        {
+            Some(consumed) if consumed < 0 => {
+                *consumed_ptr = consumed;
+                Ok(u64::try_from(-consumed).unwrap())
+            }
+            _ => bail!("not enough fuel remaining in store"),
+        }
+    }
+
     #[inline]
     pub fn signal_handler(&self) -> Option<*const SignalHandler<'static>> {
         let handler = self.signal_handler.as_ref()?;
@@ -940,15 +1145,33 @@ impl StoreInnermost {
         &*self.interrupts as *const VMInterrupts as *mut VMInterrupts
     }
 
-    pub unsafe fn insert_vmexternref(&mut self, r: VMExternRef) {
-        self.externref_activations_table
-            .insert_with_gc(r, &self.modules)
+    pub unsafe fn insert_vmexternref_without_gc(&mut self, r: VMExternRef) {
+        self.externref_activations_table.insert_without_gc(r);
     }
 
     #[inline]
     pub fn default_callee(&self) -> *mut VMContext {
         self.default_callee.vmctx_ptr()
     }
+
+    pub fn traitobj(&self) -> *mut dyn wasmtime_runtime::Store {
+        self.default_callee.store()
+    }
+
+    /// Takes the cached `Vec<Val>` stored internally across hostcalls to get
+    /// used as part of calling the host in a `Func::new` method invocation.
+    pub fn take_hostcall_val_storage(&mut self) -> Vec<Val> {
+        mem::take(&mut self.hostcall_val_storage)
+    }
+
+    /// Restores the vector previously taken by `take_hostcall_val_storage`
+    /// above back into the store, allowing it to be used in the future for the
+    /// next wasm->host call.
+    pub fn save_hostcall_val_storage(&mut self, storage: Vec<Val>) {
+        if storage.capacity() > self.hostcall_val_storage.capacity() {
+            self.hostcall_val_storage = storage;
+        }
+    }
 }
 
 impl<T> StoreContextMut<'_, T> {
@@ -1228,9 +1451,9 @@ impl AsyncCx {
                 Poll::Pending => {}
             }
 
-            let before = wasmtime_runtime::TlsRestore::take().map_err(Trap::from_runtime)?;
+            let before = wasmtime_runtime::TlsRestore::take().map_err(Trap::from_runtime_box)?;
             let res = (*suspend).suspend(());
-            before.replace().map_err(Trap::from_runtime)?;
+            before.replace().map_err(Trap::from_runtime_box)?;
             res?;
         }
     }
@@ -1238,7 +1461,7 @@ impl AsyncCx {
 
 unsafe impl<T> wasmtime_runtime::Store for StoreInner<T> {
     fn vminterrupts(&self) -> *mut VMInterrupts {
-        <StoreInnermost>::vminterrupts(self)
+        <StoreOpaque>::vminterrupts(self)
     }
 
     fn externref_activations_table(
@@ -1268,7 +1491,7 @@ unsafe impl<T> wasmtime_runtime::Store for StoreInner<T> {
                 }
                 *injection_count -= 1;
                 let fuel = *fuel_to_inject;
-                StoreContextMut(self).opaque().out_of_gas_yield(fuel)?;
+                self.out_of_gas_yield(fuel)?;
                 Ok(())
             }
             #[cfg(not(feature = "async"))]
@@ -1314,7 +1537,7 @@ impl<T> Drop for Store<T> {
     }
 }
 
-impl Drop for StoreInnermost {
+impl Drop for StoreOpaque {
     fn drop(&mut self) {
         // NB it's important that this destructor does not access `self.data`.
         // That is deallocated by `Drop for Store<T>` above.
