@@ -1,6 +1,5 @@
 use std::fmt;
-use wasmtime_environ::wasm::{EntityType, WasmFuncType};
-use wasmtime_environ::{ir, wasm};
+use wasmtime_environ::{EntityType, Global, Memory, Table, WasmFuncType, WasmType};
 use wasmtime_jit::TypeTables;
 
 pub(crate) mod matching;
@@ -18,43 +17,13 @@ pub enum Mutability {
     Var,
 }
 
-/// Limits of tables/memories where the units of the limits are defined by the
-/// table/memory types.
-///
-/// A minimum is always available but the maximum may not be present.
-#[derive(Debug, Clone, Hash, Eq, PartialEq)]
-pub struct Limits {
-    min: u32,
-    max: Option<u32>,
-}
-
-impl Limits {
-    /// Creates a new set of limits with the minimum and maximum both specified.
-    pub fn new(min: u32, max: Option<u32>) -> Limits {
-        Limits { min, max }
-    }
-
-    /// Creates a new `Limits` with the `min` specified and no maximum specified.
-    pub fn at_least(min: u32) -> Limits {
-        Limits::new(min, None)
-    }
-
-    /// Returns the minimum amount for these limits.
-    pub fn min(&self) -> u32 {
-        self.min
-    }
-
-    /// Returns the maximum amount for these limits, if specified.
-    pub fn max(&self) -> Option<u32> {
-        self.max
-    }
-}
-
 // Value Types
 
 /// A list of all possible value types in WebAssembly.
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
 pub enum ValType {
+    // NB: the ordering here is intended to match the ordering in
+    // `wasmtime_types::WasmType` to help improve codegen when converting.
     /// Signed 32 bit integer.
     I32,
     /// Signed 64 bit integer.
@@ -65,10 +34,10 @@ pub enum ValType {
     F64,
     /// A 128 bit number.
     V128,
-    /// A reference to opaque data in the Wasm instance.
-    ExternRef, /* = 128 */
     /// A reference to a Wasm function.
     FuncRef,
+    /// A reference to opaque data in the Wasm instance.
+    ExternRef,
 }
 
 impl fmt::Display for ValType {
@@ -103,40 +72,28 @@ impl ValType {
         }
     }
 
-    pub(crate) fn get_wasmtime_type(&self) -> ir::Type {
+    pub(crate) fn to_wasm_type(&self) -> WasmType {
         match self {
-            ValType::I32 => ir::types::I32,
-            ValType::I64 => ir::types::I64,
-            ValType::F32 => ir::types::F32,
-            ValType::F64 => ir::types::F64,
-            ValType::V128 => ir::types::I8X16,
-            ValType::ExternRef => wasmtime_runtime::ref_type(),
-            ValType::FuncRef => wasmtime_runtime::pointer_type(),
+            Self::I32 => WasmType::I32,
+            Self::I64 => WasmType::I64,
+            Self::F32 => WasmType::F32,
+            Self::F64 => WasmType::F64,
+            Self::V128 => WasmType::V128,
+            Self::FuncRef => WasmType::FuncRef,
+            Self::ExternRef => WasmType::ExternRef,
         }
     }
 
-    pub(crate) fn to_wasm_type(&self) -> wasm::WasmType {
-        match self {
-            Self::I32 => wasm::WasmType::I32,
-            Self::I64 => wasm::WasmType::I64,
-            Self::F32 => wasm::WasmType::F32,
-            Self::F64 => wasm::WasmType::F64,
-            Self::V128 => wasm::WasmType::V128,
-            Self::FuncRef => wasm::WasmType::FuncRef,
-            Self::ExternRef => wasm::WasmType::ExternRef,
-        }
-    }
-
-    pub(crate) fn from_wasm_type(ty: &wasm::WasmType) -> Self {
+    pub(crate) fn from_wasm_type(ty: &WasmType) -> Self {
         match ty {
-            wasm::WasmType::I32 => Self::I32,
-            wasm::WasmType::I64 => Self::I64,
-            wasm::WasmType::F32 => Self::F32,
-            wasm::WasmType::F64 => Self::F64,
-            wasm::WasmType::V128 => Self::V128,
-            wasm::WasmType::FuncRef => Self::FuncRef,
-            wasm::WasmType::ExternRef => Self::ExternRef,
-            wasm::WasmType::ExnRef => unimplemented!(),
+            WasmType::I32 => Self::I32,
+            WasmType::I64 => Self::I64,
+            WasmType::F32 => Self::F32,
+            WasmType::F64 => Self::F64,
+            WasmType::V128 => Self::V128,
+            WasmType::FuncRef => Self::FuncRef,
+            WasmType::ExternRef => Self::ExternRef,
+            WasmType::ExnRef => unimplemented!(),
         }
     }
 }
@@ -198,10 +155,7 @@ impl ExternType {
         (Instance(InstanceType) instance unwrap_instance)
     }
 
-    pub(crate) fn from_wasmtime(
-        types: &TypeTables,
-        ty: &wasmtime_environ::wasm::EntityType,
-    ) -> ExternType {
+    pub(crate) fn from_wasmtime(types: &TypeTables, ty: &EntityType) -> ExternType {
         match ty {
             EntityType::Function(idx) => {
                 FuncType::from_wasm_func_type(types.wasm_signatures[*idx].clone()).into()
@@ -217,7 +171,7 @@ impl ExternType {
                 let ty = &types.instance_signatures[*ty];
                 InstanceType::from_wasmtime(types, ty).into()
             }
-            EntityType::Event(_) => unimplemented!("wasm event support"),
+            EntityType::Tag(_) => unimplemented!("wasm tag support"),
         }
     }
 }
@@ -276,28 +230,28 @@ impl FuncType {
         results: impl IntoIterator<Item = ValType>,
     ) -> FuncType {
         FuncType {
-            sig: WasmFuncType {
-                params: params.into_iter().map(|t| t.to_wasm_type()).collect(),
-                returns: results.into_iter().map(|t| t.to_wasm_type()).collect(),
-            },
+            sig: WasmFuncType::new(
+                params.into_iter().map(|t| t.to_wasm_type()).collect(),
+                results.into_iter().map(|t| t.to_wasm_type()).collect(),
+            ),
         }
     }
 
     /// Returns the list of parameter types for this function.
     pub fn params(&self) -> impl ExactSizeIterator<Item = ValType> + '_ {
-        self.sig.params.iter().map(ValType::from_wasm_type)
+        self.sig.params().iter().map(ValType::from_wasm_type)
     }
 
     /// Returns the list of result types for this function.
     pub fn results(&self) -> impl ExactSizeIterator<Item = ValType> + '_ {
-        self.sig.returns.iter().map(ValType::from_wasm_type)
+        self.sig.returns().iter().map(ValType::from_wasm_type)
     }
 
-    pub(crate) fn as_wasm_func_type(&self) -> &wasm::WasmFuncType {
+    pub(crate) fn as_wasm_func_type(&self) -> &WasmFuncType {
         &self.sig
     }
 
-    pub(crate) fn from_wasm_func_type(sig: wasm::WasmFuncType) -> FuncType {
+    pub(crate) fn from_wasm_func_type(sig: WasmFuncType) -> FuncType {
         Self { sig }
     }
 }
@@ -337,7 +291,7 @@ impl GlobalType {
 
     /// Returns `None` if the wasmtime global has a type that we can't
     /// represent, but that should only very rarely happen and indicate a bug.
-    pub(crate) fn from_wasmtime_global(global: &wasm::Global) -> GlobalType {
+    pub(crate) fn from_wasmtime_global(global: &Global) -> GlobalType {
         let ty = ValType::from_wasm_type(&global.wasm_ty);
         let mutability = if global.mutability {
             Mutability::Var
@@ -357,38 +311,46 @@ impl GlobalType {
 /// which `call_indirect` can invoke other functions.
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
 pub struct TableType {
-    element: ValType,
-    limits: Limits,
+    ty: Table,
 }
 
 impl TableType {
     /// Creates a new table descriptor which will contain the specified
     /// `element` and have the `limits` applied to its length.
-    pub fn new(element: ValType, limits: Limits) -> TableType {
-        TableType { element, limits }
+    pub fn new(element: ValType, min: u32, max: Option<u32>) -> TableType {
+        TableType {
+            ty: Table {
+                wasm_ty: element.to_wasm_type(),
+                minimum: min,
+                maximum: max,
+            },
+        }
     }
 
     /// Returns the element value type of this table.
-    pub fn element(&self) -> &ValType {
-        &self.element
+    pub fn element(&self) -> ValType {
+        ValType::from_wasm_type(&self.ty.wasm_ty)
     }
 
-    /// Returns the limits, in units of elements, of this table.
-    pub fn limits(&self) -> &Limits {
-        &self.limits
+    /// Returns minimum number of elements this table must have
+    pub fn minimum(&self) -> u32 {
+        self.ty.minimum
     }
 
-    pub(crate) fn from_wasmtime_table(table: &wasm::Table) -> TableType {
-        let ty = match table.ty {
-            wasm::TableElementType::Func => ValType::FuncRef,
-            #[cfg(target_pointer_width = "64")]
-            wasm::TableElementType::Val(ir::types::R64) => ValType::ExternRef,
-            #[cfg(target_pointer_width = "32")]
-            wasm::TableElementType::Val(ir::types::R32) => ValType::ExternRef,
-            _ => panic!("only `funcref` and `externref` tables supported"),
-        };
-        let limits = Limits::new(table.minimum, table.maximum);
-        TableType::new(ty, limits)
+    /// Returns the optionally-specified maximum number of elements this table
+    /// can have.
+    ///
+    /// If this returns `None` then the table is not limited in size.
+    pub fn maximum(&self) -> Option<u32> {
+        self.ty.maximum
+    }
+
+    pub(crate) fn from_wasmtime_table(table: &Table) -> TableType {
+        TableType { ty: table.clone() }
+    }
+
+    pub(crate) fn wasmtime_table(&self) -> &Table {
+        &self.ty
     }
 }
 
@@ -400,23 +362,78 @@ impl TableType {
 /// chunks of addressable memory.
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
 pub struct MemoryType {
-    limits: Limits,
+    ty: Memory,
 }
 
 impl MemoryType {
-    /// Creates a new descriptor for a WebAssembly memory given the specified
-    /// limits of the memory.
-    pub fn new(limits: Limits) -> MemoryType {
-        MemoryType { limits }
+    /// Creates a new descriptor for a 32-bit WebAssembly memory given the
+    /// specified limits of the memory.
+    ///
+    /// The `minimum` and `maximum`  values here are specified in units of
+    /// WebAssembly pages, which are 64k.
+    pub fn new(minimum: u32, maximum: Option<u32>) -> MemoryType {
+        MemoryType {
+            ty: Memory {
+                memory64: false,
+                shared: false,
+                minimum: minimum.into(),
+                maximum: maximum.map(|i| i.into()),
+            },
+        }
     }
 
-    /// Returns the limits (in pages) that are configured for this memory.
-    pub fn limits(&self) -> &Limits {
-        &self.limits
+    /// Creates a new descriptor for a 64-bit WebAssembly memory given the
+    /// specified limits of the memory.
+    ///
+    /// The `minimum` and `maximum`  values here are specified in units of
+    /// WebAssembly pages, which are 64k.
+    ///
+    /// Note that 64-bit memories are part of the memory64 proposal for
+    /// WebAssembly which is not standardized yet.
+    pub fn new64(minimum: u64, maximum: Option<u64>) -> MemoryType {
+        MemoryType {
+            ty: Memory {
+                memory64: true,
+                shared: false,
+                minimum,
+                maximum,
+            },
+        }
     }
 
-    pub(crate) fn from_wasmtime_memory(memory: &wasm::Memory) -> MemoryType {
-        MemoryType::new(Limits::new(memory.minimum, memory.maximum))
+    /// Returns whether this is a 64-bit memory or not.
+    ///
+    /// Note that 64-bit memories are part of the memory64 proposal for
+    /// WebAssembly which is not standardized yet.
+    pub fn is_64(&self) -> bool {
+        self.ty.memory64
+    }
+
+    /// Returns minimum number of WebAssembly pages this memory must have.
+    ///
+    /// Note that the return value, while a `u64`, will always fit into a `u32`
+    /// for 32-bit memories.
+    pub fn minimum(&self) -> u64 {
+        self.ty.minimum
+    }
+
+    /// Returns the optionally-specified maximum number of pages this memory
+    /// can have.
+    ///
+    /// If this returns `None` then the memory is not limited in size.
+    ///
+    /// Note that the return value, while a `u64`, will always fit into a `u32`
+    /// for 32-bit memories.
+    pub fn maximum(&self) -> Option<u64> {
+        self.ty.maximum
+    }
+
+    pub(crate) fn from_wasmtime_memory(memory: &Memory) -> MemoryType {
+        MemoryType { ty: memory.clone() }
+    }
+
+    pub(crate) fn wasmtime_memory(&self) -> &Memory {
+        &self.ty
     }
 }
 
